from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.utils.deprecation import MiddlewareMixin

from ridi_django_oauth2.config import RidiOAuth2Config
from ridi_django_oauth2.response import HttpUnauthorizedResponse
from ridi_django_oauth2.utils.token import get_token_from_cookie, get_token_info
from ridi_oauth2.client.dtos import TokenData
from ridi_oauth2.introspector.dtos import AccessTokenInfo
from ridi_oauth2.introspector.exceptions import PublicKeyException


class AuthenticationMiddleware(MiddlewareMixin):
    def process_request(self, request):
        request.user = AnonymousUser()

        token = get_token_from_cookie(request=request)
        token_info = None
        if token.access_token:
            try:
                token_info = get_token_info(token.access_token.token)
            except PublicKeyException:
                return HttpUnauthorizedResponse()

        if token_info is not None:
            self._set_user_in_request(request, token_info, token)

        return None

    @staticmethod
    def _set_user_in_request(request, token_info: AccessTokenInfo, token: TokenData):
        get_user_from_token_info = RidiOAuth2Config.get_user_from_token_info_callable()

        if get_user_from_token_info:
            user = get_user_from_token_info(token_info)

        else:
            user, _ = get_user_model().objects.get_or_create(u_idx=token_info.u_idx)

        user.token = token
        user.token_info = token_info
        request.user = user





class OAuth2GrantType:
    AUTHORIZATION_CODE = 'authorization_code'
    IMPLICIT = 'implicit'
    REOUSRCE_OWNER_PASSWORD_CREDENTIALS = 'resource_owner_password_credentials'
    CLIENT_CREDENTIALS = 'client_credentials'
    REFRESH_TOKEN = 'refresh_token'


class OAuth2ErrorCode:
    INVALID_REQUEST = 'invalid_request'
    INVALID_CLIENT = 'invalid_client'
    INVALID_GRANT = 'invalid_grant'
    UNAUTHORIZED_CLIENT = 'unauthorized_client'
    UNSUPPORTED_GRANT_TYPE = 'unsupported_grant_type'
    INVALID_SCOPE = 'invalid_scope'



# Generated by Django 2.0.2 on 2018-03-19 11:43

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='RidiUser',
            fields=[
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('u_idx', models.AutoField(editable=False, primary_key=True, serialize=False, verbose_name='u_idx')),
            ],
            options={
                'verbose_name': '사용자 계정',
                'verbose_name_plural': '사용자 계정 리스트',
                'db_table': 'ridi_user',
            },
        ),
    ]



from typing import Dict


class ClientInfo:
    def __init__(self, client_id: str, client_secret: str, scope: str = None, redirect_uri: str = None):
        self._client_id = client_id
        self._client_secret = client_secret
        self._scope = scope
        self._redirect_uri = redirect_uri

    @property
    def client_id(self) -> str:
        return self._client_id

    @property
    def client_secret(self) -> str:
        return self._client_secret

    @property
    def scope(self) -> str:
        return self._scope

    @property
    def redirect_uri(self) -> str:
        return self._redirect_uri


class AuthorizationServerInfo:
    def __init__(self, authorization_url: str = None, token_url: str = None):
        self._authorization_url = authorization_url
        self._token_url = token_url

    @property
    def authorization_url(self) -> str:
        return self._authorization_url

    @property
    def token_url(self) -> str:
        return self._token_url


class Token:
    def __init__(self, token: str, expires_in: int = None):
        self._token = token
        self._expires_in = expires_in

    @property
    def token(self) -> str:
        return self._token

    @property
    def expires_in(self) -> int:
        return self._expires_in


class TokenData:
    def __init__(self, access_token: Token, token_type: str = None, scope: str = None, refresh_token: Token = None):
        self._access_token = access_token
        self._token_type = token_type
        self._scope = scope

        self._refresh_token = refresh_token

    @property
    def access_token(self) -> Token:
        return self._access_token

    @property
    def token_type(self) -> str:
        return self._token_type

    @property
    def scope(self) -> str:
        return self._scope

    @property
    def refresh_token(self) -> Token:
        return self._refresh_token

    @staticmethod
    def from_dict(dictionary: Dict):
        access_token = None
        if dictionary.get('access_token', None):
            access_token = Token(token=dictionary['access_token'], expires_in=dictionary.get('expires_in', None))

        refresh_token = None
        if dictionary.get('refresh_token', None):
            refresh_token = Token(token=dictionary['refresh_token'], expires_in=dictionary.get('refresh_token_expires_in', None))

        return TokenData(
            access_token=access_token, token_type=dictionary.get('token_type', None), scope=dictionary.get('scope', None),
            refresh_token=refresh_token
        )



from typing import Dict, List

import requests
from requests import RequestException, Response

from ridi_django_oauth2_lib.decorators.retry import RetryFailException, retry
from ridi_django_oauth2.config import RidiOAuth2Config
from ridi_oauth2.introspector.constants import JWKKeyType, JWKUse
from ridi_oauth2.introspector.dtos import JWKDto
from ridi_oauth2.introspector.exceptions import AccountServerException, ClientRequestException, FailToLoadPublicKeyException, \
    InvalidPublicKey, NotExistedKey


class KeyHandler:
    _public_key_dtos = {}

    @classmethod
    def _get_memorized_key_dto(cls, client_id: str, kid: str) -> JWKDto:
        return cls._public_key_dtos.get(client_id, {}).get(kid, None)

    @classmethod
    def get_public_key_by_kid(cls, client_id: str, kid: str):
        public_key_dto = cls._get_memorized_key_dto(client_id, kid)

        if not public_key_dto or public_key_dto.is_expired:
            public_key_dto = cls._reset_key_dtos(client_id, kid)

        cls._assert_valid_key(public_key_dto)

        return public_key_dto.public_key

    @staticmethod
    def _assert_valid_key(key: JWKDto):
        if not key:
            raise NotExistedKey
        if key.kty != JWKKeyType.RSA or key.use != JWKUse.SIG:
            raise InvalidPublicKey

    @classmethod
    def _reset_key_dtos(cls, client_id: str, kid: str) -> JWKDto:
        try:
            keys = cls._get_valid_public_keys_by_client_id(client_id)

        except RetryFailException:
            raise FailToLoadPublicKeyException

        cls._memorize_key_dtos(client_id, keys)

        return cls._get_memorized_key_dto(client_id, kid)

    @classmethod
    def _memorize_key_dtos(cls, client_id: str, keys: List[JWKDto]):
        key_dtos = cls._public_key_dtos.get(client_id, {})
        for key in keys:
            key_dtos[key.kid] = key
        cls._public_key_dtos[client_id] = key_dtos

    @staticmethod
    def _process_response(response: Response) -> Dict:
        if response.status_code >= 500:
            raise AccountServerException
        elif response.status_code >= 400:
            raise ClientRequestException
        return response.json()

    @classmethod
    @retry(retry_count=3, retriable_exceptions=(RequestException, AccountServerException,))
    def _get_valid_public_keys_by_client_id(cls, client_id: str) -> List[JWKDto]:
        response = requests.request(
            method='GET',
            url=RidiOAuth2Config.get_key_url(),
            params={'client_id': client_id},
        )
        return [JWKDto(key) for key in cls._process_response(response=response).get('keys')]




import random
import string as string_lib

UNICODE_ASCII_CHARACTER_SET = string_lib.ascii_letters + string_lib.digits
DEFAULT_LENGTH = 32


def generate_random_str(length=DEFAULT_LENGTH, chars=UNICODE_ASCII_CHARACTER_SET):
    return ''.join(random.choices(chars, k=length))



import typing

from ridi_oauth2.resource.constants import Scope

REQUIRE_SCOPE_TYPE = typing.TypeVar('REQUIRE_SCOPE_TYPE', typing.Tuple, typing.List, str)


class ScopeChecker:
    @classmethod
    def check(cls, require_scopes: typing.List[REQUIRE_SCOPE_TYPE], user_scopes: typing.List[str]) -> bool:
        if Scope.ALL in user_scopes:
            return True

        if not require_scopes:
            return True

        return cls._or(require_scopes=require_scopes, user_scopes=user_scopes)

    @classmethod
    def _or(cls, require_scopes: typing.List[REQUIRE_SCOPE_TYPE], user_scopes: typing.List[str]) -> bool:
        return any(cls._iterate_scopes(require_scopes=require_scopes, user_scopes=user_scopes))

    @staticmethod
    def _and(and_scopes: typing.Iterable, user_scopes: typing.List[str]) -> bool:
        for scope in and_scopes:
            if scope not in user_scopes:
                return False

        return True

    @classmethod
    def _iterate_scopes(cls, require_scopes: typing.List[REQUIRE_SCOPE_TYPE], user_scopes: typing.List[str]) -> typing.Iterable[bool]:
        for require_scope in require_scopes:
            if isinstance(require_scope, str):
                yield require_scope in user_scopes
            elif isinstance(require_scope, typing.Iterable):  # check isinstance tuple or list
                yield cls._and(and_scopes=require_scope, user_scopes=user_scopes)
            else:
                yield False



import jwt
from jwt import InvalidTokenError
from jwt.exceptions import InvalidKeyError

from ridi_oauth2.introspector.dtos import AccessTokenInfo
from ridi_oauth2.introspector.exceptions import InvalidJwtSignatureException, InvalidToken
from ridi_oauth2.introspector.key_handler import KeyHandler


class JwtIntrospectHelper:
    @staticmethod
    def introspect(access_token: str) -> AccessTokenInfo:
        try:
            unverified_header = jwt.get_unverified_header(access_token)
            unverified_payload = jwt.decode(access_token, verify=False)
        except jwt.InvalidTokenError:
            raise InvalidToken

        kid = unverified_header.get('kid', None)
        client_id = unverified_payload.get('client_id', None)

        if not kid or not client_id:
            raise InvalidJwtSignatureException

        public_key = KeyHandler.get_public_key_by_kid(client_id, kid)

        try:
            payload = jwt.decode(jwt=access_token, key=public_key, algorithms=unverified_header.get('alg'))
            return AccessTokenInfo.from_dict(payload)
        except (InvalidTokenError, InvalidKeyError):
            raise InvalidJwtSignatureException
